;; Generated package description from tramp.el  -*- no-byte-compile: t -*-
(define-package "tramp" "2.6.2.2" "Transparent Remote Access, Multiple Protocol" '((emacs "26.1")) :commit "cab3f20dcab56c116d0694ce6a5c2985616ae044" :authors '(("Kai Großjohann" . "kai.grossjohann@gmx.net")) :maintainer '("Michael Albinus" . "michael.albinus@gmx.de") :keywords '("comm" "processes") :url "https://www.gnu.org/software/tramp/")
